# Copyright 2020 Kazys Stepanas
# Licensed under the Apache License, Version 2.0

from colcon_cmake.task.cmake import _parse_cmake_version_string


def test_parse_cmake_version():
    # Build version prefix string closely matching what cmake version outputs.
    base_prefix = 'cmake version '

    # Expected results list. Each element is a tuple containing the following:
    # - Version string to parse.
    # - Numeric version tuple to compare against (major, minor, patch).
    # The second item is None where the parse string should not parse.
    test_items = [
        (base_prefix + '3.0.0', (3, 0, 0)),
        (base_prefix + '3.0.0-dirty', (3, 0, 0)),
        (base_prefix + '3.0.0-rc1', (3, 0, 0)),
        (base_prefix + 'cmake version 3.0.0-rc1-dirty', (3, 0, 0)),
        (base_prefix + 'this.is.garbage', None),
        (base_prefix + '3.15.1', (3, 15, 1)),
        ('3.15.1', (3, 15, 1)),
        (base_prefix + '101.202.303-xxx', (101, 202, 303)),
        ('101.202.303-xxx', (101, 202, 303)),
        ('prefix 1 number 101.202.303-xxx', (101, 202, 303)),
        ('not the right format', None)
    ]

    # Iterate the strings and parse.
    for version_string, expected_version in test_items:
        parsed_version = _parse_cmake_version_string(version_string)
        if expected_version is None:
            # Input string was garbage. Assert parsing failed.
            assert parsed_version is None
        else:
            assert parsed_version._version.release[0:3] == expected_version


if __name__ == '__main__':
    test_parse_cmake_version()



# Copyright 2016-2018 Dirk Thomas
# Copyright 2024 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_cmake.package_identification.cmake \
    import CmakePackageIdentification
from colcon_core.package_descriptor import PackageDescriptor


def test_identify():
    extension = CmakePackageIdentification()

    with TemporaryDirectory(prefix='test_colcon_') as basepath:
        desc = PackageDescriptor(basepath)
        desc.type = 'other'
        assert extension.identify(desc) is None
        assert desc.name is None

        desc.type = None
        assert extension.identify(desc) is None
        assert desc.name is None
        assert desc.type is None

        basepath = Path(basepath)
        (basepath / 'CMakeLists.txt').write_text('')
        assert extension.identify(desc) is None
        assert desc.name == basepath.name
        assert desc.type == 'cmake'

        desc = PackageDescriptor(basepath)
        (basepath / 'CMakeLists.txt').write_text(
            'cmake_minimum_required(VERSION 3.10)\n'
            'project(Project NONE)\n')
        assert extension.identify(desc) is None
        assert desc.name == 'Project'
        assert desc.type == 'cmake'

        desc = PackageDescriptor(basepath)
        (basepath / 'CMakeLists.txt').write_text(
            'cmake_minimum_required(VERSION 3.10)\n'
            'project(Project NONE)\n'
            'catkin_workspace()\n')
        assert extension.identify(desc) is None
        assert desc.name is None
        assert desc.type is None

        desc = PackageDescriptor(basepath)
        (basepath / 'CMakeLists.txt').write_text(
            'cmake_minimum_required(VERSION 3.10)\n'
            'project(pkg-name NONE)\n')
        assert extension.identify(desc) is None
        assert desc.name == 'pkg-name'
        assert desc.type == 'cmake'
        assert set(desc.dependencies.keys()) == {'build', 'run'}
        assert not desc.dependencies['build']
        assert not desc.dependencies['run']
        assert extension.identify(desc) is None
        assert desc.name == 'pkg-name'
        assert desc.type == 'cmake'

        desc = PackageDescriptor(basepath)
        (basepath / 'CMakeLists.txt').write_text(
            'cmake_minimum_required(VERSION 3.10)\n'
            'project(other-name NONE)\n'
            'find_package(PkgConfig REQUIRED)\n'
            'pkg_check_modules(DEP_NAME REQUIRED dep-name>=1.1)\n'
            'add_subdirectory(src)\n')
        (basepath / 'src').mkdir(parents=True, exist_ok=True)
        (basepath / 'src' / 'CMakeLists.txt').write_text(
            'find_package(dep-name2 REQUIRED)\n')
        (basepath / 'src' / 'README.txt').write_text(
            'find_package(other-dep-name REQUIRED)\n')
        assert extension.identify(desc) is None
        assert desc.name == 'other-name'
        assert desc.type == 'cmake'
        assert set(desc.dependencies.keys()) == {'build', 'run'}
        assert desc.dependencies['build'] == {
            'dep-name', 'dep-name2', 'PkgConfig',
        }
        assert desc.dependencies['run'] == {
            'dep-name', 'dep-name2', 'PkgConfig',
        }



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
import re

from colcon_core.dependency_descriptor import DependencyDescriptor
from colcon_core.logging import colcon_logger
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.plugin_system import satisfies_version

logger = colcon_logger.getChild(__name__)


class CmakePackageIdentification(PackageIdentificationExtensionPoint):
    """Identify CMake packages with `CMakeLists.txt` files."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def identify(self, metadata):  # noqa: D102
        if metadata.type is not None and metadata.type != 'cmake':
            return

        cmakelists_txt = metadata.path / 'CMakeLists.txt'
        if not cmakelists_txt.is_file():
            return

        data = extract_data(cmakelists_txt)
        if not data['name'] and not metadata.name:
            raise RuntimeError(
                "Failed to extract project name from '%s'" % cmakelists_txt)
        if data['name'] == 'Project':
            lines = cmakelists_txt.read_text(errors='replace').splitlines()
            if 'catkin_workspace()' in lines:
                logger.warning(
                    f"Ignoring '{cmakelists_txt}' since it seems to be a "
                    "toplevel CMake file generated by 'catkin_make'")
                return

        metadata.type = 'cmake'
        if metadata.name is None:
            metadata.name = data['name']
        metadata.dependencies['build'] |= data['depends']
        metadata.dependencies['run'] |= data['depends']


def extract_data(cmakelists_txt):
    """
    Extract the project name and dependencies from a CMakeLists.txt file.

    :param Path cmakelists_txt: The path of the CMakeLists.txt file
    :rtype: dict
    """
    content = extract_content(cmakelists_txt)

    data = {}
    data['name'] = extract_project_name(content)
    # fall back to use the directory name
    if data['name'] is None:
        data['name'] = cmakelists_txt.parent.name

    # extract dependencies from all CMake files in the project directory
    depends_content = content + extract_content(
        cmakelists_txt.parent, exclude=[cmakelists_txt])

    depends = extract_dependencies(depends_content)
    # exclude self references
    data['depends'] = depends - {data['name']}

    return data


def extract_content(basepath, exclude=None):
    """
    Get all non-comment lines from CMake files under the given basepath.

    All `CMakeLists.txt` files as well as files ending with `.cmake` are used.
    Directories starting with a dot (`.`) are being skipped.

    :param Path basepath: The path to recursively crawl
    :param list exclude: The paths to exclude
    :rtype: str
    """
    if basepath.is_file():
        content = basepath.read_text(errors='replace')
    elif basepath.is_dir():
        content = ''
        for dirpath, dirnames, filenames in os.walk(str(basepath)):
            # skip subdirectories starting with a dot
            dirnames[:] = filter(lambda d: not d.startswith('.'), dirnames)
            dirnames.sort()

            for name in sorted(filenames):
                if name != 'CMakeLists.txt' and not name.endswith('.cmake'):
                    continue

                path = Path(dirpath) / name
                if path in (exclude or []):
                    continue

                content += path.read_text(errors='replace') + '\n'
    else:
        return ''
    return _remove_cmake_comments(content)


def _remove_cmake_comments(content):
    lines = content.splitlines()
    for index, line in enumerate(lines):
        lines[index] = _remove_cmake_comments_from_line(line)
    return '\n'.join(lines)


def _remove_cmake_comments_from_line(line):
    # match comments starting with #
    # which are not within a string enclosed in double quotes
    pattern = (
        # strings
        '("[^"]*")'
        '|'
        # comments
        '(#.*)'
        '|'
        # other
        '([^#"]*)'
    )

    modline = ''
    for matches in re.findall(pattern, line):
        modline += matches[0] + matches[2]
    return modline


def extract_project_name(content):
    """
    Extract the CMake project name from the CMake code.

    The `project()` call must be on a single line and the first argument must
    be a literal string for this function to be able to extract the name.

    :param str content: The CMake source code
    :returns: The project name, otherwise None
    :rtype: str
    """
    # extract project name
    match = re.search(
        # case insensitive function name
        '(?i:project)'
        # optional white space
        r'\s*'
        # open parenthesis
        r'\('
        # optional white space
        r'\s*'
        # optional "opening" quote
        '("?)'
        # project name
        '([a-zA-Z0-9_-]+)'
        # optional "closing" quote (only if an "opening" quote was used)
        r'\1'
        # optional language
        r'(\s+[^\)]*)?'
        # close parenthesis
        r'\)',
        content)
    if not match:
        return None
    return match.group(2)


def extract_dependencies(content):
    """
    Extract the dependencies from the CMake code.

    The `find_package()` and `pkg_check_modules` calls must be on a single line
    and the first argument must be a literal string for this function to be
    able to extract the dependency name.

    :param str content: The CMake source code
    :returns: The dependencies name
    :rtype: list
    """
    metadata = {
        'origin': 'cmake',
    }

    return {
        DependencyDescriptor(d, metadata=metadata)
        for d in extract_find_package_calls(content) |
        _extract_pkg_config_calls(content)
    }


def extract_find_package_calls(content, *, function_name='find_package'):
    """
    Extract `find_package()`-like calls from the CMake code.

    The function call must be on a single line and the first argument must be
    a literal string for this function to be able to extract it.

    :param str content: The CMake source code
    :returns: The first arguments of the functions (in the case of
      `find_package()` these are the package names)
    :rtype: set
    """
    matches = re.findall(
        # case insensitive function name
        f'(?i:{function_name})'
        # optional white space
        r'\s*'
        # open parenthesis
        r'\('
        # optional white space
        r'\s*'
        # optional "opening" quote
        '("?)'
        # package name
        '([a-zA-Z0-9_-]+)'
        # optional "closing" quote (only if an "opening" quote was used)
        r'\1'
        # white space
        r'(\s+'
        # optional arguments
        r'[^\)]*)?'
        # close parenthesis
        r'\)',
        content)
    return {m[1] for m in matches}


def _extract_pkg_config_calls(content):
    pattern1 = '(?i:pkg_check_modules)'
    pattern2 = '(?i:pkg_search_module)'
    function_names_pattern = f'(?:{pattern1}|{pattern2})'
    matches = re.findall(
        # case insensitive function names
        function_names_pattern +
        # optional white space
        r'\s*'
        # open parenthesis
        r'\('
        # optional white space
        r'\s*'
        # optional "opening" quote
        '("?)'
        # prefix
        '[a-zA-Z0-9_]+'
        # optional "closing" quote (only if an "opening" quote was used)
        r'\1'
        # optional options prefixed by white space
        r'(?:\s+(?:REQUIRED|QUIET|NO_CMAKE_PATH|NO_CMAKE_ENVIRONMENT_PATH))*'
        # package names prefixed by white space with opt. trailing white space
        '([^)]+)'
        # close parenthesis
        r'\)',
        content)
    names = set()
    for modules in [m[1].strip() for m in matches]:
        # split multiple modules
        for module in modules.split():
            # remove optional version suffix
            for char in ('>', '=', '<'):
                if char in module:
                    module = module[:module.index(char)]
            names.add(module)
    return names



# Copyright 2019 Rover Robotics
# Copyright 2024 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

import asyncio
import os
from pathlib import Path
import shutil
from types import SimpleNamespace

from colcon_cmake.task.cmake.build import CmakeBuildTask
from colcon_core.event_handler.console_direct import ConsoleDirectEventHandler
from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.subprocess import new_event_loop
from colcon_core.task import TaskContext
import pytest


@pytest.fixture(autouse=True)
def monkey_patch_put_event_into_queue(monkeypatch):
    event_handler = ConsoleDirectEventHandler()
    monkeypatch.setattr(
        TaskContext,
        'put_event_into_queue',
        lambda self, event: event_handler((event, 'cmake')),
    )


def _test_build_package(
    tmp_path, *, cmake_args=None, cmake_clean_cache=False,
    cmake_clean_first=False, cmake_force_configure=None,
    cmake_target=None, cmake_target_skip_unavailable=None,
):
    event_loop = new_event_loop()
    asyncio.set_event_loop(event_loop)

    try:
        package = PackageDescriptor(tmp_path / 'src')
        package.name = 'test-package'
        package.type = 'cmake'

        context = TaskContext(
            pkg=package,
            args=SimpleNamespace(
                path=str(tmp_path / 'src'),
                build_base=str(tmp_path / 'build'),
                install_base=str(tmp_path / 'install'),
                cmake_args=cmake_args,
                cmake_clean_cache=cmake_clean_cache,
                cmake_clean_first=cmake_clean_first,
                cmake_force_configure=cmake_force_configure,
                cmake_target=cmake_target,
                cmake_target_skip_unavailable=cmake_target_skip_unavailable,
            ),
            dependencies={}
        )

        task = CmakeBuildTask()
        task.set_context(context=context)

        package.path.mkdir(exist_ok=True)
        (package.path / 'CMakeLists.txt').write_text(
            'cmake_minimum_required(VERSION 3.5)\n'
            'project(test-package NONE)\n'
            'file(GENERATE OUTPUT "${CMAKE_CURRENT_BINARY_DIR}/out-file"\n'
            '     CONTENT "Hello, World!")\n'
            'install(FILES "${CMAKE_CURRENT_BINARY_DIR}/out-file"\n'
            '        DESTINATION "share")\n'
            'add_custom_target(custom-target\n'
            '                  ${CMAKE_COMMAND} -E echo "Hello, World!")\n'
        )

        src_base = Path(task.context.args.path)

        source_files_before = set(src_base.rglob('*'))
        rc = event_loop.run_until_complete(task.build())
        assert not rc
        source_files_after = set(src_base.rglob('*'))
        assert source_files_before == source_files_after

        install_base = Path(task.context.args.install_base)
        assert (cmake_target is None) == \
            (install_base / 'share' / 'out-file').is_file()
    finally:
        event_loop.close()


@pytest.mark.parametrize(
    'cmake_target',
    [None, 'custom-target'])
@pytest.mark.skipif(
    not shutil.which('cmake'),
    reason='CMake must be installed to run this test')
@pytest.mark.skipif(
    os.name == 'nt' and 'VisualStudioVersion' not in os.environ,
    reason='Must be run from a developer command prompt')
def test_build_package(tmpdir, cmake_target):
    tmp_path = Path(tmpdir)
    _test_build_package(tmp_path, cmake_target=cmake_target)



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from xml.etree.ElementTree import ElementTree

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_test_result.test_result import Result
from colcon_test_result.test_result import TestResultExtensionPoint

logger = colcon_logger.getChild(__name__)


class CtestTestResult(TestResultExtensionPoint):
    """
    Collect the CTest results generated when testing a set of CMake packages.

    It checks each direct subdirectory of the passed build base for a
    'Testing/TAG' file.
    The first line in that file contains the directory name of the latest
    'Test.xml' result file relative to the 'Testing' directory.
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            TestResultExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def get_test_results(  # noqa: D102
        self, basepath, *, collect_details, files=None
    ):
        results = set()
        # check all 'TAG' files in a directory named 'Testing'
        for tag_file in basepath.glob('**/Testing/TAG'):
            if not tag_file.is_file():
                continue

            if files is not None:
                files.add(str(tag_file))

            # find the latest Test.xml file
            latest_xml_dir = tag_file.read_text().splitlines()[0]
            latest_xml_path = tag_file.parent / latest_xml_dir / 'Test.xml'
            if not latest_xml_path.exists():
                logger.warning(
                    f"Skipping '{tag_file}': could not find latest XML file "
                    f"'{latest_xml_path}'")
                continue

            # parse the XML file
            tree = ElementTree()
            root = tree.parse(str(latest_xml_path))

            # check if the root tag looks like a CTest file
            if root.tag != 'Site':
                logger.warning(
                    f"Skipping '{latest_xml_path}': the root tag is not "
                    "'Site'")
                continue

            # look for a single 'Testing' child tag
            children = list(root)
            if len(children) != 1:
                logger.warning(
                    f"Skipping '{latest_xml_path}': 'Site' tag is expected to "
                    '"have exactly one child')
                continue
            if children[0].tag != 'Testing':
                logger.warning(
                    f"Skipping '{latest_xml_path}': the child tag is not "
                    "'Testing'")
                continue

            if files is not None:
                files.add(str(latest_xml_path))

            # collect information from all 'Test' tags
            result = Result(str(latest_xml_path))
            for child in children[0]:
                if child.tag != 'Test':
                    continue

                result.test_count += 1

                try:
                    status = child.attrib['Status']
                except KeyError:
                    logger.warning(
                        f"Skipping '{latest_xml_path}': a 'test' tag lacks a "
                        "'Status' attribute")
                    break

                if status == 'failed':
                    result.failure_count += 1
                elif status == 'notrun':
                    result.skipped_count += 1
                elif status != 'passed':
                    result.error_count += 1

                if collect_details and status == 'failed':
                    lines = [child.findtext('Name')]
                    lines.extend(
                        _get_messages(
                            'failure message',
                            child.findtext('Results/Measurement/Value')))
                    result.details.append('\n'.join(lines))
            else:
                results.add(result)
        return results


def _get_messages(label, message):
    lines = []
    if message:
        lines.append('<<< ' + label)
        for line in message.strip('\n\r').splitlines():
            lines.append('  ' + line)
        lines.append('>>>')
    return lines



# Copyright 2016-2018 Dirk Thomas
# Copyright 2024 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

from colcon_cmake.environment.cmake_prefix_path \
    import CmakePrefixPathEnvironment


def test_cmake_prefix_path():
    extension = CmakePrefixPathEnvironment()

    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)
        with patch(
            'colcon_cmake.environment.cmake_prefix_path'
            '.create_environment_hook',
            return_value=['/some/hook', '/other/hook']
        ):
            # No CMake configs exist
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            pkg_share_path = prefix_path / 'share' / 'pkg_name'

            # Unrelated file
            unrelated_file = pkg_share_path / 'cmake' / 'README.md'
            unrelated_file.parent.mkdir(parents=True, exist_ok=True)
            unrelated_file.touch()
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            # PkgNameConfig.cmake exists
            cmake_config = pkg_share_path / 'cmake' / 'PkgNameConfig.cmake'
            cmake_config.parent.mkdir(parents=True, exist_ok=True)
            cmake_config.touch()
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 2
            cmake_config.unlink()

            # pkg_name-config.cmake exists
            cmake_config = pkg_share_path / 'cmake' / 'pkg_name-config.cmake'
            cmake_config.parent.mkdir(parents=True, exist_ok=True)
            cmake_config.touch()
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 2
            cmake_config.unlink()



# Copyright 2016-2018 Dirk Thomas
# Copyright 2024 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

from colcon_cmake.environment.cmake_module_path \
    import CmakeModulePathEnvironment


def test_cmake_module_path():
    extension = CmakeModulePathEnvironment()

    with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
        prefix_path = Path(prefix_path)
        with patch(
            'colcon_cmake.environment.cmake_module_path.'
            'create_environment_hook',
            return_value=['/some/hook', '/other/hook']
        ):
            # No CMake configs exist
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            pkg_share_path = prefix_path / 'share' / 'pkg_name'

            # Unrelated file
            unrelated_file = pkg_share_path / 'cmake' / 'README.md'
            unrelated_file.parent.mkdir(parents=True, exist_ok=True)
            unrelated_file.touch()
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 0

            # FindPkgName.cmake exists
            cmake_module = pkg_share_path / 'cmake' / 'FindPkgName.cmake'
            cmake_module.parent.mkdir(parents=True, exist_ok=True)
            cmake_module.touch()
            hooks = extension.create_environment_hooks(prefix_path, 'pkg_name')
            assert len(hooks) == 2



# Copyright 2020 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path

from colcon_cmake.task.cmake import CMAKE_EXECUTABLE
from colcon_core.event.command import Command
from colcon_core.event.job import JobQueued
from colcon_core.event.job import JobUnselected
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_reactor import EventReactorShutdown
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version

logger = colcon_logger.getChild(__name__)


class CompileCommandsEventHandler(EventHandlerExtensionPoint):
    """
    Generate a `compile_commands.json` file for the whole workspace.

    The file is created in the build directory and aggregates the data from all
    packages.
    """

    FILENAME = 'compile_commands.json'

    # the priority should be lower than e.g. the status and summary extensions
    # in order to show those as soon as possible
    PRIORITY = 40

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._package_names = set()
        self._any_cmake_commands = False

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, (JobQueued, JobUnselected)):
            # delay loading json for all packages
            self._package_names.add(data.identifier)

        elif isinstance(data, Command):
            if data.cmd[0] == CMAKE_EXECUTABLE:
                self._any_cmake_commands = True

        elif isinstance(data, EventReactorShutdown):
            # if CMake was never invoked there is no need (re-)generate a
            # workspace-level json file
            if not self._any_cmake_commands:
                return

            # if no package has a json file there is no need for a
            # workspace-level json file
            package_level_json_paths = self._get_package_level_json_paths()
            workspace_level_json_path = self._get_path()
            if not package_level_json_paths:
                if workspace_level_json_path.exists():
                    workspace_level_json_path.unlink()
                return

            # if the workspace-level json file is newer than all package-level
            # json files it doesn't need to be regenerated
            try:
                workspace_level_mtime = os.path.getmtime(
                    str(workspace_level_json_path))
            except OSError:
                pass
            else:
                for json_path in sorted(package_level_json_paths):
                    try:
                        mtime = os.path.getmtime(str(json_path))
                    except OSError:
                        continue
                    if mtime > workspace_level_mtime:
                        break
                else:
                    return

            # read all package-level json data
            # and combine them in the workspace-level file
            # for performance reasons avoid to load/dump json data
            with workspace_level_json_path.open('wb') as h:
                h.write(b'[')
                # keep deterministic order independent of aborted/selected pkgs
                written_compile_commands = False
                for json_path in sorted(package_level_json_paths):
                    compile_commands = self._get_compile_commands(json_path)
                    if not compile_commands:
                        continue
                    if written_compile_commands:
                        # add an empty line between packages
                        h.write(b',\n')
                    else:
                        written_compile_commands = True
                    h.write(b'\n')
                    h.write(compile_commands)
                if written_compile_commands:
                    h.write(b'\n')
                h.write(b']\n')

    def _get_package_level_json_paths(self):
        json_paths = set()
        for package_name in self._package_names:
            json_path = self._get_path(package_name)
            if json_path.exists():
                json_paths.add(json_path)
        return json_paths

    def _get_compile_commands(self, json_path):
        content = json_path.read_bytes()
        try:
            open_index = content.index(b'[')
            close_index = content.rindex(b']')
        except ValueError:
            logger.warning(
                "Data in '%s' is expected to be a list" % json_path.absolute())
            return None
        return content[open_index + 1:close_index - 1].strip()

    def _get_path(self, package_name=None):
        path = Path(self.context.args.build_base)
        if package_name is not None:
            path /= package_name
        path /= CompileCommandsEventHandler.FILENAME
        return path
